# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import tempfile
import unittest
from dataclasses import replace
from pathlib import Path
from textwrap import dedent
from typing import Optional

from ..api import usort, usort_path
from ..config import Config
from ..translate import import_from_node
from ..util import parse_import

DEFAULT_CONFIG = Config()


class BasicOrderingTest(unittest.TestCase):
    maxDiff = None

    def test_order(self) -> None:
        items_in_order = [
            "from __future__ import division",
            "import os",
            "from os import path",
            "import tp",
            "from tp import x",
            "from .. import c",
            "from . import a",
            "from . import b",
            "from .a import z",
        ]

        nodes = [
            import_from_node(parse_import(x), config=DEFAULT_CONFIG)
            for x in items_in_order
        ]
        self.assertSequenceEqual(nodes, sorted(nodes))


class UsortStringFunctionalTest(unittest.TestCase):
    def assertUsortResult(
        self, before: str, after: str, config: Optional[Config] = None
    ) -> None:
        before = dedent(before)
        after = dedent(after)
        config = config or DEFAULT_CONFIG
        result1 = usort(before.encode(), config)  # first pass
        result2 = usort(result1.output, config)  # enforce stable sorting on second pass
        if result1.error:
            raise result1.error
        if result2.error:
            raise result2.error
        if result2.output != result1.output:
            self.fail(
                "µsort result was not stable on second pass:\n\n"
                f"Before:\n-------\n{before}\n"
                f"First Pass:\n-----------\n{result1.output.decode()}\n"
                f"Second Pass:\n------------\n{result2.output.decode()}"
            )
        if result2.output.decode() != after:
            self.fail(
                "µsort result did not match expected value:\n\n"
                f"Before:\n-------\n{before}\n"
                f"Expected:\n---------\n{after}\n"
                f"Result:\n-------\n{result1.output.decode()}"
            )

    def test_sort_ordering(self) -> None:
        # This only tests ordering, not any of the comment or whitespace
        # modifications.
        self.assertUsortResult(
            """
                import a
                import a.b
                from a.b import foo2
                from a import foo
                import b
            """,
            """
                import a
                import a.b
                import b
                from a import foo
                from a.b import foo2
            """,
        )

    def test_sort_blocks(self) -> None:
        # This only tests that there are two blocks and we only reorder within a
        # block
        self.assertUsortResult(
            """
                import d
                import c
                print("hi")
                import b
                import a
            """,
            """
                import c
                import d
                print("hi")
                import a
                import b
            """,
        )

    # Disabled until wrapping is supported
    #     def test_sort_wrap_moves_comments(self):
    #         # Test that end-of-line directive comments get moved to the first line
    #         # when wrapping is going to happen.
    #         self.assertEqual(
    #             """\
    # from a (  # pylint: disable=E1
    #     import foo,
    # )
    # """,
    #             usort_string(
    #                 """\
    # from a import foo  # pylint: disable=E1
    # """,
    #                 line_limit=10,
    #             ),
    #         )

    def test_star_imports(self) -> None:
        # Test that we create a second block with the star import
        self.assertUsortResult(
            """
                import d
                import c
                from x import *
                import b
                import a
            """,
            """
                import c
                import d
                from x import *
                import a
                import b
            """,
        )

    def test_shadowed_import(self) -> None:
        # Test that a new block is started when there's a duplicate name
        self.assertUsortResult(
            """
                import b as b
                import a as b
            """,
            """
                import b as b
                import a as b
            """,
        )

    def test_shadowed_import_ok(self) -> None:
        self.assertUsortResult(
            """
                import a.d
                import a.c
                import a.b
            """,
            """
                import a.b
                import a.c
                import a.d
            """,
        )

    def test_shadowed_relative_import_ok(self) -> None:
        self.assertUsortResult(
            """
                from os import path as path
                from os import path
                import os.path as path
            """,
            """
                import os.path as path
                from os import path, path as path
            """,
        )

    def test_semicolons(self) -> None:
        self.assertUsortResult(
            """
                import zipfile
                import re
                import sys; print(sys)
                import urllib
                import difflib
            """,
            """
                import re
                import zipfile
                import sys; print(sys)
                import difflib
                import urllib
            """,
        )

    def test_dot_handling(self) -> None:
        # Test that 'from .. import b' comes before 'from ..a import foo'
        self.assertUsortResult(
            """
                from ..a import foo
                from .. import b
                from . import d
                from fp import z
                import fp
                from .c import e
            """,
            """
                import fp
                from fp import z

                from .. import b
                from ..a import foo
                from . import d
                from .c import e
            """,
        )

    def test_customized_sections(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort]
categories = ["future", "standard_library", "numpy", "third_party", "first_party"]
[tool.usort.known]
numpy = ["numpy", "pandas"]
"""
            )
            sample = Path(d) / "sample.py"
            conf = Config.find(sample)
            self.assertUsortResult(
                """
                    import os
                    from . import foo
                    import numpy as np
                    import aaa
                """,
                """
                    import os

                    import numpy as np

                    import aaa

                    from . import foo
                """,
                conf,
            )

    def test_non_module_imports(self) -> None:
        self.assertUsortResult(
            """
                if True:
                    import b
                    import a

                def func():
                    import d
                    import c
                    if True:
                        import f
                        import e
                        pass
                        import a
            """,
            """
                if True:
                    import a
                    import b

                def func():
                    import c
                    import d
                    if True:
                        import e
                        import f
                        pass
                        import a
            """,
        )

    def test_whitespace_between_sections(self) -> None:
        self.assertUsortResult(
            """
                from __future__ import unicode_literals
                from __future__ import division
                import sys



                import third_party
                #comment
                from . import first_party
            """,
            """
                from __future__ import division, unicode_literals

                import sys

                import third_party

                #comment
                from . import first_party
            """,
        )

    def test_whitespace_between_sections_indented(self) -> None:
        self.assertUsortResult(
            """
                import sys
                def foo():
                    import os
                    from something import nothing
            """,
            # TODO: maybe we should add whitespace after the last import in a block?
            """
                import sys
                def foo():
                    import os

                    from something import nothing
            """,
        )

    def test_case_insensitive_sorting(self) -> None:
        content = """
            import calendar
            import cProfile
            import dataclasses

            from fissix.main import diff_texts
            from IPython import start_ipython
            from libcst import Module
        """
        self.assertUsortResult(content, content)

    def test_side_effect_modules(self) -> None:
        config = replace(
            DEFAULT_CONFIG,
            side_effect_modules=["tracemalloc", "fizzbuzz", "foo.bar.baz"],
        )
        content = """
            from zipfile import ZipFile
            from tracemalloc import start
            from collections import defaultdict

            import fizzbuzz
            import attr
            import solar
            import foo.bar.baz
            from foo import bar
            from star import sun
            from foo.bar import baz
            from attr import evolve
        """
        self.assertUsortResult(content, content, config)

    def test_match_black_blank_line_before_comment(self) -> None:
        content = """
            import a
            import b

            # comment
            import c
        """
        self.assertUsortResult(content, content)

    def test_multi_line_comments(self) -> None:
        self.assertUsortResult(
            """
                from fuzz import buzz
                # one
                from foo import (  # two
                    # three
                    beta,  # four
                    # five
                    gamma, # six
                    alpha # seven
                    , # eight
                    # nine
                )  # ten
                # eleven
            """,
            """
                # one
                from foo import (  # two
                    alpha,  # seven  # eight
                    # three
                    beta,  # four
                    # five
                    gamma,  # six
                    # nine
                )  # ten
                from fuzz import buzz
                # eleven
            """,
        )

    def test_multi_line_maintain(self) -> None:
        self.assertUsortResult(
            """
                from phi import rho
                # one
                from alpha import (  # two
                    # three
                    beta,  # four
                    # five
                    delta # six
                    , # seven
                    gamma # eight
                    # nine
                )  # ten
                # eleven
            """,
            """
                # one
                from alpha import (  # two
                    # three
                    beta,  # four
                    # five
                    delta,  # six  # seven
                    gamma,  # eight
                    # nine
                )  # ten
                from phi import rho
                # eleven
            """,
        )

    def test_multi_line_maintain_inner(self) -> None:
        self.assertUsortResult(
            """
                def foo():
                    from phi import rho
                    # one
                    from alpha import (  # two
                        # three
                        beta,  # four
                        # five
                        delta # six
                        , # seven
                        gamma # eight
                        # nine
                    )  # ten
                    # eleven
            """,
            """
                def foo():
                    # one
                    from alpha import (  # two
                        # three
                        beta,  # four
                        # five
                        delta,  # six  # seven
                        gamma,  # eight
                        # nine
                    )  # ten
                    from phi import rho
                    # eleven
            """,
        )

    def test_multi_line_collapse(self) -> None:
        self.assertUsortResult(
            """
                from fuzz import buzz
                # 1
                from a import (  # 2
                    # 3
                    b,  # 4
                    # 5
                    c # 6
                    , # 7
                    d # 8
                    # 9
                )  # 10
                # 11
            """,
            """
                # 1
                from a import b, c, d  # 2  # 3  # 4  # 5  # 6  # 7  # 8  # 9  # 10
                from fuzz import buzz
                # 11
            """,
        )

    def test_multi_line_collapse_inner(self) -> None:
        self.assertUsortResult(
            """
                def foo():
                    from fuzz import buzz
                    # 1
                    from foo import (  # 2
                        # 3
                        bar,  # 4
                        # 5
                        baz # 6
                        , # 7
                    )  # 8
                    # 9
            """,
            """
                def foo():
                    # 1
                    from foo import bar, baz  # 2  # 3  # 4  # 5  # 6  # 7  # 8
                    from fuzz import buzz
                    # 9
            """,
        )

    def test_maintain_tabs(self) -> None:
        self.assertUsortResult(
            """
                import foo

                def a():
                \timport bar
                \t
                \tdef b():
                \t\timport baz
            """,
            """
                import foo

                def a():
                \timport bar
                \t
                \tdef b():
                \t\timport baz
            """,
        )

    def test_multi_line_expand_top_level(self) -> None:
        self.assertUsortResult(
            """
                from really_absurdly_long_python_module_name.insanely_long_submodule_name import SomeReallyObnoxiousCamelCaseClass
            """,
            """
                from really_absurdly_long_python_module_name.insanely_long_submodule_name import (
                    SomeReallyObnoxiousCamelCaseClass,
                )
            """,
        )

    def test_multi_line_expand_function(self) -> None:
        self.assertUsortResult(
            """
                def foo():
                    from really_absurdly_long_python_module_name.insanely_long_submodule_name import SomeReallyObnoxiousCamelCaseClass
            """,
            """
                def foo():
                    from really_absurdly_long_python_module_name.insanely_long_submodule_name import (
                        SomeReallyObnoxiousCamelCaseClass,
                    )
            """,
        )

    def test_multi_line_expand_but_not_very_long(self) -> None:
        """
        This test makes sure indentation is taken into account properly.
        Without indentation, these lines wouldn't be too long.
        """
        self.assertUsortResult(
            """
                if foo:
                    from aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa import bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb
            """,
            """
                if foo:
                    from aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa import (
                        bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb,
                    )
            """,
        )

    def test_multi_line_expand_inner_function(self) -> None:
        self.assertUsortResult(
            """
                def foo():
                    import b
                    import a

                    def bar():
                        from really_absurdly_long_python_module_name.insanely_long_submodule_name import SomeReallyObnoxiousCamelCaseClass
            """,
            """
                def foo():
                    import a
                    import b

                    def bar():
                        from really_absurdly_long_python_module_name.insanely_long_submodule_name import (
                            SomeReallyObnoxiousCamelCaseClass,
                        )
            """,
        )

    def test_single_line_parens(self) -> None:
        self.assertUsortResult(
            """
                from delta import echo, foxtrot
                from alfa import (bravo, charlie)  # hello
            """,
            """
                from alfa import bravo, charlie  # hello
                from delta import echo, foxtrot
            """,
        )

    def test_single_line_import_long_comment(self) -> None:
        """
        Basic import statements can't be reflowed to multiple lines
        """
        self.assertUsortResult(
            """
                import foo, bar, baz  # some really long inline comment that would can't be reflowed to a new line
            """,
            """
                import bar, baz, foo  # some really long inline comment that would can't be reflowed to a new line
            """,
        )

    def test_sorting_import_items(self) -> None:
        self.assertUsortResult(
            """
                import b, a, c
                from typing import List, Dict, Set, Optional, Pattern
            """,
            """
                from typing import Dict, List, Optional, Pattern, Set

                import a, b, c
            """,
        )

    def test_sorting_import_items_comments(self) -> None:
        self.assertUsortResult(
            """
                # zero
                from foo import (  # one
                    # two
                    gamma,  # three
                    bravo,  # four
                    delta,  # five
                    alpha,  # six
                    # seven
                )  # eight
            """,
            """
                # zero
                from foo import (  # one
                    alpha,  # six
                    bravo,  # four
                    delta,  # five
                    # two
                    gamma,  # three
                    # seven
                )  # eight
            """,
        )

    def test_sorting_mixed_category_imports(self) -> None:
        self.assertUsortResult(
            """
                import os, attr
                import difflib
                import third_party
            """,
            """
                import difflib

                import attr, os
                import third_party
            """,
        )

    def test_merging_import_items(self) -> None:
        self.assertUsortResult(
            """
                import os
                import os.path
                from typing import List, Dict, Tuple, Set, Optional
                import os
                from typing import Union, Sequence
                from foo.bar import b, d, c
                from foo import baz
                from foo.bar import c as C, a, d
                from foo import fizz
            """,
            """
                import os
                import os.path
                from typing import Dict, List, Optional, Sequence, Set, Tuple, Union

                from foo import baz, fizz
                from foo.bar import a, b, c, c as C, d
            """,
        )

    def test_merging_imports_with_multiple_blocks(self) -> None:
        """verify that merging doesn't affect future blocks/statements"""
        self.assertUsortResult(
            """
                import os
                import os.path
                from datetime import date
                from datetime import datetime
                from typing import *
                import re
                from pathlib import Path
                print("hello world")
            """,
            """
                import os
                import os.path
                from datetime import date, datetime
                from typing import *
                import re
                from pathlib import Path
                print("hello world")
            """,
        )

    def test_merging_import_items_comments(self) -> None:
        self.assertUsortResult(
            """
                import a

                # one
                from foo import ( # two
                    # three
                    beta,  # four
                    # five
                    delta,  # six
                    # seven
                )  # eight
                # apple
                from foo import (  # banana
                    # cranberry
                    gamma,  # date
                    # elderberry
                    alpha,  # fig
                    # grape
                    beta,  # hazelnut
                    # kiwi
                ) # lime
                # mango
            """,
            """
                import a

                # one
                # apple
                from foo import (  # two  # banana
                    # elderberry
                    alpha,  # fig
                    # three
                    # grape
                    beta,  # four  # hazelnut
                    # five
                    delta,  # six
                    # cranberry
                    gamma,  # date
                    # seven
                    # kiwi
                )  # eight  # lime
                # mango
            """,
        )

    def test_merging_imports_disabled(self) -> None:
        self.assertUsortResult(
            """
                import os
                import os.path
                from typing import List, Dict, Tuple, Set, Optional
                import os
                from typing import Union, Sequence
                from foo.bar import b, d, c
                from foo import baz
                from foo.bar import c as C, a, d
                from foo import fizz
            """,
            """
                import os
                import os
                import os.path
                from typing import Dict, List, Optional, Set, Tuple
                from typing import Sequence, Union

                from foo import baz
                from foo import fizz
                from foo.bar import a, c as C, d
                from foo.bar import b, c, d
            """,
            Config(merge_imports=False),
        )

    def test_sort_implicit_blocks1(self) -> None:
        self.assertUsortResult(
            """
                from phi import phi
                from alpha import SHADOW
                from delta import delta
                from eta import eta
                from mu import SHADOW
                from chi import chi
                from beta import beta
            """,
            """
                from alpha import SHADOW
                from beta import beta
                from chi import chi
                from delta import delta
                from eta import eta
                from mu import SHADOW
                from phi import phi
            """,
        )

    def test_sort_implicit_blocks2(self) -> None:
        self.assertUsortResult(
            """
                from mu import SHADOW
                from delta import delta
                from eta import eta
                from chi import chi
                from alpha import SHADOW
                from beta import beta
            """,
            """
                from chi import chi
                from delta import delta
                from eta import eta
                from mu import SHADOW
                from alpha import SHADOW
                from beta import beta
            """,
        )

    def test_sort_implicit_blocks3(self) -> None:
        self.assertUsortResult(
            """
                from phi import phi
                from delta import SHADOW
                from eta import eta
                from chi import chi
                from alpha import SHADOW
                from beta import beta
            """,
            """
                from chi import chi
                from delta import SHADOW
                from alpha import SHADOW
                from beta import beta
                from eta import eta
                from phi import phi
            """,
        )

    def test_skip_directives(self) -> None:
        """Test both usort:skip and isort:skip on single line imports"""
        self.assertUsortResult(
            """
                from os import path
                import functools  # usort: skip
                import asyncio
                from collections import defaultdict  # isort:skip
                from asyncio import gather
            """,
            """
                from os import path
                import functools  # usort: skip
                import asyncio
                from collections import defaultdict  # isort:skip
                from asyncio import gather
            """,
        )

    def test_skip_directives_after_noqa(self) -> None:
        """Test that skips are obeyed, even if they aren't the first directive"""
        self.assertUsortResult(
            """
                from os import path
                import functools  # noqa  # usort: skip
                import asyncio
            """,
            """
                from os import path
                import functools  # noqa  # usort: skip
                import asyncio
            """,
        )

    def test_skip_directives_multiline(self) -> None:
        """Validate that skip work on both first and last line of multiline imports"""
        self.assertUsortResult(
            """
                from unittest.mock import (
                    Mock, MagicMock, call, patch, sentinel, ANY,
                )  # usort:skip
                from functools import wraps
                from asyncio import (  # usort:skip
                    gather, wait,
                )
                from collections import defaultdict
            """,
            """
                from unittest.mock import (
                    Mock, MagicMock, call, patch, sentinel, ANY,
                )  # usort:skip
                from functools import wraps
                from asyncio import (  # usort:skip
                    gather, wait,
                )
                from collections import defaultdict
            """,
        )

    def test_excludes(self) -> None:
        original_content = "import os\nimport asyncio\n"
        sorted_content = b"import asyncio\nimport os\n"
        with tempfile.TemporaryDirectory() as td:
            tdp = Path(td).resolve()
            (tdp / "pyproject.toml").write_text(
                """\
[tool.usort]
excludes = [
    "fixtures/",
    "*generated.py",
]
"""
            )
            (tdp / "foo" / "tests" / "fixtures").mkdir(parents=True)

            excluded_paths = (
                (tdp / "foo" / "tests" / "fixtures" / "foo.py"),
                (tdp / "foo" / "tests" / "foo_generated.py"),
                (tdp / "foo" / "client_generated.py"),
            )
            sorted_paths = (
                (tdp / "foo" / "tests" / "foo.py"),
                (tdp / "foo" / "foo.py"),
                (tdp / "foo" / "generated_client.py"),
            )

            for path in excluded_paths + sorted_paths:
                path.write_text(original_content)

            results = list(usort_path(tdp / "foo"))
            for result in results:
                self.assertIn(result.path, sorted_paths)
                self.assertNotIn(result.path, excluded_paths)
                self.assertEqual(sorted_content, result.output.replace(b"\r\n", b"\n"))
            self.assertEqual(len(sorted_paths), len(results))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

import datetime

project = "µsort"
copyright = f"{datetime.date.today().year}, Meta Platforms, Inc. and affiliates."
author = "Tim Hatch"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_mdinclude",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
}
autodoc_member_order = "groupwise"

highlight_language = "python3"
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pathspec": ("https://python-path-specification.readthedocs.io/en/latest/", None),
}
master_doc = "index"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"
html_theme_options = {
    "description": "Safe Python import sorting",
    "fixed_sidebar": True,
    "badge_branch": "master",
    "github_button": False,
    "github_user": "facebook",
    "github_repo": "usort",
    "show_powered_by": False,
    "sidebar_collapse": False,
    "extra_nav_links": {
        "Report Issues": "https://github.com/facebook/usort/issues",
    },
}

html_sidebars = {
    "**": [
        "about.html",
        "badges.html",
        "navigation.html",
        "relations.html",
        "searchbox.html",
        "meta.html",
    ],
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Set, Tuple

import libcst as cst
from libcst.metadata import PositionProvider

from .config import Config
from .translate import import_from_node, import_to_node
from .types import SortableBlock, SortableImport, SortWarning
from .util import timed

LOG = logging.getLogger(__name__)


class ImportSorter:
    def __init__(self, *, module: cst.Module, path: Path, config: Config):
        self.config = config
        self.module = module
        self.path = path
        self.warning_nodes: List[Tuple[cst.CSTNode, str]] = []
        self.warnings: List[SortWarning] = []
        self.wrapper = cst.MetadataWrapper(module, unsafe_skip_copy=True)
        self.transformer = ImportSortingTransformer(config, module, self)

    def has_skip_comment(self, comment: Optional[cst.Comment]) -> bool:
        if comment is None or not comment.value:
            return False

        directives = comment.value.split("#")
        for directive in directives:
            directive = directive.replace(" ", "")
            if directive.startswith("usort:skip") or directive.startswith("isort:skip"):
                return True

        return False

    def is_sortable_import(self, stmt: cst.CSTNode) -> bool:
        """
        Determine if any individual statement is sortable or should be a barrier.

        Handles skip directives, configured side effect modules, and star imports.
        """
        if isinstance(stmt, cst.SimpleStatementLine):
            # from foo import (
            #     bar,
            # )  # usort:skip
            comment = stmt.trailing_whitespace.comment
            if self.has_skip_comment(comment):
                return False

            # N.b. `body` is a list, because the SimpleStatementLine might have
            # semicolons. We treat lines with multiple statements as unsortable.
            # This is the safest and easiest thing to do.
            #
            # If black is run, it will put the semicolon pieces on different lines,
            # but we don't want to do that until we reflow and handle directives
            # like noqa.  TODO do that before calling is_sortable_import, and assert
            # that it's not a compound statement line.
            if len(stmt.body) > 1:
                return False
            elif isinstance(stmt.body[0], cst.ImportFrom):
                # from foo import (  # usort:skip
                #     bar,
                # )
                if isinstance(stmt.body[0].lpar, cst.LeftParen) and isinstance(
                    stmt.body[0].lpar.whitespace_after, cst.ParenthesizedWhitespace
                ):
                    comment = stmt.body[0].lpar.whitespace_after.first_line.comment
                    if self.has_skip_comment(comment):
                        return False
                # `from x import *` is a barrier
                if isinstance(stmt.body[0].names, cst.ImportStar):
                    return False
                # check for side effect modules, but ignore local (from .) imports (TODO?)
                elif stmt.body[0].module is not None:
                    base = cst.helpers.get_full_name_for_node_or_raise(
                        stmt.body[0].module
                    )
                    names = [name.evaluated_name for name in stmt.body[0].names]
                    if self.config.is_side_effect_import(base, names):
                        return False
                return True
            elif isinstance(stmt.body[0], cst.Import):
                base = ""
                names = [name.evaluated_name for name in stmt.body[0].names]
                if self.config.is_side_effect_import(base, names):
                    return False
                return True
            else:
                return False
        else:
            return False

    def name_overlap(self, block: SortableBlock, imp: SortableImport) -> Set[str]:
        """
        Find imported names that overlap existing names for a block of imports.

        Compares imports of a proposed, but not yet included, import with existing imports
        in a block. Ignores multiple imports of the same "name" that come from the same
        qualified name. Eg, `os.path` doesn't shadow `os`, but `from foo import os` does.

        Returns a set of qualified names from `imp` that shadow names from `block`.
        This set will be empty if there are no overlaps.
        """
        overlap: Set[str] = set()

        for key, value in imp.imported_names.items():
            shadowed = block.imported_names.get(key)
            if shadowed and shadowed != value:
                self.warning_nodes.append(
                    (
                        imp.node,
                        f"Name {shadowed!r} shadowed by {value!r}; "
                        "implicit block split",
                    )
                )
                overlap.add(shadowed)

        return overlap

    def split_inplace(self, block: SortableBlock, overlap: Set[str]) -> SortableBlock:
        """
        Split an existing block into two blocks after the last shadowed import.

        Pre-sorts the block of imports, then finds the last import with shadowed names, and
        splits after that import. Returns a new block containing all imports after the split
        point, or empty otherwise.
        """
        # best-effort pre-sorting before we split
        for imp in block.imports:
            imp.items.sort()
        block.imports.sort()

        # find index of last shadowed import, starting from the end of the block's imports
        idx = len(block.imports)
        while idx > 0:
            idx -= 1
            imp = block.imports[idx]
            if any(item.fullname in overlap for item in imp.items):
                break

        count = idx + 1
        if count >= len(block.imports):
            # shadowed import is the last import in the block, so we can't split anything.
            # return a new, empty block following pattern from sortable_blocks()
            new = SortableBlock(block.end_idx, block.end_idx + 1)

        else:
            # Split the existing block after the shadowed import, creating a new block that
            # starts after the shadowed import, update the old block's end index, and then
            # move all the imports after that to the new block
            new = SortableBlock(block.start_idx + count, block.end_idx)
            block.end_idx = block.start_idx + count

            new.imports = block.imports[count:]
            block.imports[count:] = []

            # move imported names metadata
            for imp in new.imports:
                for key in list(imp.imported_names):
                    new.imported_names[key] = block.imported_names.pop(key)

        return new

    def sortable_blocks(self, body: Sequence[cst.BaseStatement]) -> List[SortableBlock]:
        """
        Finds blocks of imports separated by barriers (non-import statements, or
        dangerous imports).  We will only sort within a block, and only when there
        are no duplicate names.
        """
        blocks: List[SortableBlock] = []
        current: Optional[SortableBlock] = None
        for idx, stmt in enumerate(body):
            if self.is_sortable_import(stmt):
                assert isinstance(stmt, cst.SimpleStatementLine)
                imp = import_from_node(stmt, self.config)
                if current is None:
                    current = SortableBlock(idx, idx + 1)
                    blocks.append(current)

                overlap = self.name_overlap(current, imp)
                if overlap:
                    # This overwrites an earlier name
                    current = self.split_inplace(current, overlap)
                    blocks.append(current)

                current.add_import(imp, idx)
            else:
                current = None
        return blocks

    def partition_leading_lines(
        self,
        lines: List[str],
    ) -> Tuple[List[str], List[str]]:
        """
        Returns a tuple of the initial blank lines, and the comment lines.
        """
        for j in range(len(lines)):
            if lines[j].startswith("#"):
                break
        else:
            j = len(lines)

        return lines[:j], lines[j:]

    def fixup_whitespace(
        self, initial_blank: Sequence[str], imports: List[SortableImport]
    ) -> List[SortableImport]:
        """
        Normalize whitespace/comments on a block of imports before transforming back to CST.
        """
        cur_category = None
        # TODO if they've already been reshuffled, there may have been a blank
        # (separator) line between a non-block and the first import, that's now in
        # the middle.
        for imp in imports:
            _old_blanks, old_comments = self.partition_leading_lines(
                imp.comments.before
            )

            if cur_category is None:
                blanks = initial_blank
            elif imp.sort_key.category_index != cur_category:
                blanks = ("",)
            else:
                blanks = _old_blanks[:1]

            imp.comments.before = [*blanks, *old_comments]

            cur_category = imp.sort_key.category_index
        return imports

    def merge_and_sort_imports(
        self, imports: List[SortableImport]
    ) -> List[SortableImport]:
        if self.config.merge_imports:
            # Look for sequential imports with matching stems, and merge them.
            idx = 0
            while idx + 1 < len(imports):
                imp = imports[idx]
                nxt = imports[idx + 1]

                if (
                    # This is a from-import and the next statement is from the same module
                    (imp.stem and imp.stem == nxt.stem)
                    # This is a module-import and the next statement imports the same module
                    or (imp.stem is None and imp.items == nxt.items)
                ):
                    # Merge them and discard the second import from the list.
                    imports[idx] += nxt
                    imports.pop(idx + 1)

                else:
                    idx += 1

        # Sort items within each remaining statement
        for imp in imports:
            imp.items.sort()
            imp.calculate_sort_key()

        return imports

    def find_and_sort_blocks(
        self,
        body: Sequence[cst.BaseStatement],
        module: cst.Module,
        indent: str,
    ) -> Sequence[cst.BaseStatement]:
        """
        Find all sortable blocks in a module, sort them, and return updated module content.
        """
        sorted_body: List[cst.BaseStatement] = list(body)
        blocks = list(self.sortable_blocks(body))

        for block in blocks:
            initial_blank, initial_comment = self.partition_leading_lines(
                block.imports[0].comments.before
            )
            block.imports[0].comments.before = initial_comment
            # Sort the imports first, so that imports from the same module line up, then
            # merge and sort imports/items, then re-sort the final set of imports again
            # in case unsorted items affected overall sorting.
            imports = sorted(block.imports)
            imports = self.merge_and_sort_imports(imports)
            imports = self.fixup_whitespace(initial_blank, imports)
            block.imports = sorted(imports)

        # replace statements in reverse order in case some got merged, which throws off
        # indexes for statements past the merge
        for block in reversed(blocks):
            sorted_body[block.start_idx : block.end_idx] = [
                import_to_node(imp, module, indent, self.config)
                for imp in block.imports
            ]

        return sorted_body

    def sort_module(self) -> cst.Module:
        with timed(f"sorting {self.path}"):
            new_module = self.wrapper.visit(self.transformer)
            if self.warning_nodes:
                positions = self.wrapper.resolve(PositionProvider)
                self.warnings = [
                    SortWarning(
                        positions[self.transformer.get_original_node(node)].start.line,
                        msg,
                    )
                    for (node, msg) in self.warning_nodes
                ]
            return new_module


class ImportSortingTransformer(cst.CSTTransformer):
    def __init__(
        self, config: Config, module: cst.Module, sorter: ImportSorter
    ) -> None:
        self.config = config
        self.module = module
        self.sorter = sorter
        self.statement_map: Dict[cst.CSTNode, cst.SimpleStatementLine] = {}
        self.default_indent: str = module.default_indent
        self.indent: str = ""

    def get_original_node(self, node: cst.CSTNode) -> cst.CSTNode:
        return self.statement_map.get(node, node)

    def on_visit(self, node: cst.CSTNode) -> bool:
        if isinstance(node, (cst.BaseExpression, cst.BaseSmallStatement)):
            return False
        return super().on_visit(node)

    def leave_SimpleStatementLine(
        self,
        original_node: cst.SimpleStatementLine,
        updated_node: cst.SimpleStatementLine,
    ) -> cst.BaseStatement:
        self.statement_map[updated_node] = original_node
        return updated_node

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        sorted_body = self.sorter.find_and_sort_blocks(
            updated_node.body, module=self.module, indent=""
        )
        return updated_node.with_changes(body=sorted_body)

    def visit_IndentedBlock(self, node: cst.IndentedBlock) -> Optional[bool]:
        node_indent = node.indent
        self.indent += self.default_indent if node_indent is None else node_indent
        return True

    def leave_IndentedBlock(
        self, original_node: cst.IndentedBlock, updated_node: cst.IndentedBlock
    ) -> cst.BaseSuite:
        node_indent = original_node.indent
        if node_indent is None:
            node_indent = self.default_indent
        sorted_body = self.sorter.find_and_sort_blocks(
            updated_node.body, module=self.module, indent=self.indent
        )
        self.indent = self.indent[: -len(node_indent)]
        return updated_node.with_changes(body=sorted_body)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .cli import CliTest
from .config import ConfigTest
from .functional import BasicOrderingTest, UsortStringFunctionalTest
from .stdlibs import StdlibsTest
from .translate import IsSortableTest, SortableImportTest
from .types import TypesTest
from .util import UtilTest

__all__ = [
    "CliTest",
    "ConfigTest",
    "BasicOrderingTest",
    "UsortStringFunctionalTest",
    "IsSortableTest",
    "SortableImportTest",
    "StdlibsTest",
    "TypesTest",
    "UtilTest",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Check the current repo against all previous supported versions of µsort

If any previous version triggers sorting/formatting changes, we consider that
worthy of a major version bump, and should be reverted or at least reconsidered.

See the Versioning Guide for details.

TODO: include other projects as part of testing, like black, libcst, etc
"""

import json
import platform
import subprocess
import sys
import venv
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional
from urllib.request import urlopen

from packaging.version import Version

REPO_ROOT = Path(__file__).parent.resolve()
MINIMUM_VERSION = Version("1.0.0")
PYPI_JSON_URL = "https://pypi.org/pypi/usort/json"


def get_current_version() -> Version:
    with TemporaryDirectory() as td:
        root = Path(td).resolve()
        usort = setup_virtualenv(root)
        proc = subprocess.run(
            (usort, "--version"), encoding="utf-8", capture_output=True, check=True
        )
        return Version(proc.stdout.rpartition(" ")[-1])


def get_public_versions(
    current_version: Version, minimum_version: Version
) -> List[Version]:
    """
    Find all non-yanked versions of usort.

    Limits results such that TARGET_VERSION <= CANDIDATE_VERSION <= CURRENT_VERSION
    """
    with urlopen(PYPI_JSON_URL) as request:
        data = json.loads(request.read())

    versions: List[Version] = []
    for version_str in data["releases"]:
        version = Version(version_str)
        if all(dist["yanked"] for dist in data["releases"][version_str]):
            continue
        if minimum_version <= version <= current_version:
            versions.append(version)

    return sorted(versions, reverse=True)


def setup_virtualenv(root: Path, version: Optional[Version] = None) -> Path:
    """
    Create venv, install usort, return path to `usort` binary
    """

    venv_path = root / f"venv-{version}" if version else root / "venv-local"
    venv.create(venv_path, clear=True, with_pip=True)

    bin_dir = (
        venv_path / "Scripts" if platform.system() == "Windows" else venv_path / "bin"
    )
    python = bin_dir / "python"

    subprocess.run((python, "-m", "pip", "-q", "install", "-U", "pip"), check=True)
    if version:
        subprocess.run(
            (python, "-m", "pip", "-q", "install", "-U", f"usort=={version}"),
            check=True,
        )
    else:
        subprocess.run(
            (python, "-m", "pip", "-q", "install", "-U", REPO_ROOT), check=True
        )
    return bin_dir / "usort"


def check_versions(versions: List[Version]) -> List[Version]:
    """
    Format with local version, then check all released versions for regressions
    """
    with TemporaryDirectory() as td:
        root = Path(td).resolve()

        try:
            print("sorting with local version ...")
            usort = setup_virtualenv(root)
            subprocess.run((usort, "--version"), check=True)
            subprocess.run((usort, "format", "usort"), check=True)
            print("done\n")
        except Exception as e:
            return [("local", e)]

        failures: List[Version] = []
        for version in versions:
            try:
                print(f"checking version {version} ...")
                usort = setup_virtualenv(root, version)
                subprocess.run((usort, "--version"), check=True)
                subprocess.run((usort, "check", "usort"), check=True)
                print("clean\n")
            except Exception as e:
                failures.append((version, e))

        return failures


def main() -> None:
    current_version = get_current_version()
    minimum_version = MINIMUM_VERSION
    print(f"{current_version = !s}\n{minimum_version = !s}\n")

    versions = get_public_versions(current_version, minimum_version)
    versions_str = ", ".join(str(v) for v in versions)
    print(f"discovered versions {versions_str}\n")

    failures = check_versions(versions)
    if failures:
        print("Sorting failed in versions:")
        for version, exc in failures:
            print(f"  {version}: {exc}")
        sys.exit(1)
    else:
        print("success!")


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from typing import Sequence

import libcst as cst

from .. import util


class UtilTest(unittest.TestCase):
    def test_parse_import_simple(self) -> None:
        node = util.parse_import("import a")
        self.assertEqual(
            cst.ensure_type(
                cst.ensure_type(
                    cst.ensure_type(node, cst.SimpleStatementLine).body[0],
                    cst.Import,
                ).names[0],
                cst.ImportAlias,
            ).name.value,
            "a",
        )

    def test_parse_import_from(self) -> None:
        node = util.parse_import("from a import x")
        inner = cst.ensure_type(
            cst.ensure_type(node, cst.SimpleStatementLine).body[0],
            cst.ImportFrom,
        )
        self.assertEqual(cst.ensure_type(inner.module, cst.Name).value, "a")
        assert isinstance(inner.names, Sequence), f"{type(inner.names)} is not Sequence"
        name = inner.names[0]
        self.assertEqual(
            cst.ensure_type(
                name,
                cst.ImportAlias,
            ).name.value,
            "x",
        )

    def test_parse_import_not_an_import(self) -> None:
        with self.assertRaisesRegex(ValueError, "not an import"):
            util.parse_import("print('hello')")

    def test_parse_import_not_a_statement(self) -> None:
        with self.assertRaisesRegex(ValueError, "not a statement"):
            util.parse_import("for foo in bar: print(foo)")

    def test_parse_import_bad_syntax(self) -> None:
        with self.assertRaises(cst.ParserSyntaxError):
            util.parse_import("say 'hello'")

    def test_split_inline_comment(self) -> None:
        self.assertEqual(
            ["# foo", "# bar"], util.split_inline_comment("blah  # foo  # bar\n")
        )

    def test_split_relative(self) -> None:
        self.assertEqual(("foo", 0), util.split_relative("foo"))
        self.assertEqual(("foo", 1), util.split_relative(".foo"))
        self.assertEqual(("foo", 2), util.split_relative("..foo"))
        self.assertEqual(("foo.bar", 0), util.split_relative("foo.bar"))
        self.assertEqual(("foo.bar", 1), util.split_relative(".foo.bar"))

    def test_stem_join(self) -> None:
        self.assertEqual("foo", util.stem_join(None, "foo"))
        self.assertEqual("foo.bar", util.stem_join("foo", "bar"))
        self.assertEqual("foo.bar.baz", util.stem_join("foo", "bar.baz"))
        self.assertEqual(".bar", util.stem_join(".", "bar"))
        self.assertEqual(".foo.bar", util.stem_join(".foo", "bar"))

    def test_top_level_name(self) -> None:
        self.assertEqual("foo", util.top_level_name("foo"))
        self.assertEqual("foo", util.top_level_name("foo.bar"))
        self.assertEqual("foo", util.top_level_name("foo.bar.baz"))
        self.assertEqual("", util.top_level_name(".foo"))

    def test_with_dots(self) -> None:
        self.assertEqual("foo", util.with_dots(cst.Name(value="foo")))
        self.assertEqual(
            "foo.bar.baz",
            util.with_dots(
                cst.Attribute(
                    value=cst.Attribute(
                        value=cst.Name("foo"),
                        attr=cst.Name("bar"),
                    ),
                    attr=cst.Name("baz"),
                )
            ),
        )

        with self.assertRaisesRegex(TypeError, "Can't with_dots"):
            util.with_dots("foo.bar")  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import sys
from functools import wraps
from pathlib import Path
from typing import Any, Callable, List, Sequence

import click
from moreorless.click import echo_color_unified_diff

from usort.translate import render_node

from . import __version__
from .api import usort_path, usort_stdin
from .config import Config
from .sorting import ImportSorter
from .types import Options
from .util import get_timings, print_timings, Timing, try_parse

BENCHMARK = False


def print_benchmark(timings: Sequence[Timing]) -> None:
    if BENCHMARK:
        print_timings(click.echo, timings=timings)


def usort_command(fn: Callable[..., int]) -> Callable[..., None]:
    """
    Run wrapped command, print timings if --benchmark, and exit with return code
    """

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        exit_code = fn(*args, **kwargs) or 0
        if BENCHMARK:
            print_timings(click.echo, timings=get_timings())
        sys.exit(exit_code)

    return wrapper


@click.group()
@click.pass_context
@click.version_option(__version__, "--version", "-V")
@click.option("--debug", is_flag=True, help="Enable debug output")
@click.option("--benchmark", is_flag=True, help="Output benchmark timing info")
def main(ctx: click.Context, benchmark: bool, debug: bool) -> None:
    global BENCHMARK
    BENCHMARK = benchmark

    level = logging.DEBUG if debug else logging.WARNING
    logging.basicConfig(level=level, stream=sys.stderr)

    ctx.obj = Options(debug=debug)


@main.command()
@click.option("--multiples", is_flag=True, help="Only show files with multiple blocks")
@click.option("--debug", is_flag=True, help="Show internal information")
@click.argument("filenames", nargs=-1)
@usort_command
def list_imports(multiples: bool, debug: bool, filenames: List[str]) -> int:
    """
    Troubleshoot sorting behavior and show import blocks
    """
    # This is used to debug the sort keys on the various lines, and understand
    # where the barriers are that produce different blocks.

    for f in filenames:
        path = Path(f)
        config = Config.find(path)
        mod = try_parse(path)
        try:
            sorter = ImportSorter(module=mod, path=path, config=config)
            blocks = sorter.sortable_blocks(mod.body)
        except Exception as e:
            print("Exception", f, e)
            continue

        if multiples and len(blocks) < 2:
            continue

        click.secho(f"{f} {len(blocks)} blocks:", fg="yellow")
        for b in blocks:
            print(f"  body[{b.start_idx}:{b.end_idx}]")
            sorted_imports = sorted(b.imports)
            if debug:
                for imp in b.imports:
                    print(
                        f"    {sorted_imports.index(imp)} {imp} "
                        f"({imp.config.category(imp.stem or '')})"
                    )
            else:
                print("Formatted:")
                print("[[[")
                for imp in sorted_imports:
                    assert imp.node is not None
                    print(render_node(imp.node), end="")
                print("]]]")

    return 0


@main.command()
@click.argument("filenames", nargs=-1)
@usort_command
def check(filenames: List[str]) -> int:
    """
    Check imports for one or more path
    """
    if not filenames:
        raise click.ClickException("Provide some filenames")

    return_code = 0
    paths = [Path(f) for f in filenames]
    for result in usort_path(paths, write=False):
        if result.error:
            click.echo(f"Error sorting {result.path}: {result.error}")
            return_code |= 1

        for warning in result.warnings:
            click.echo(f"Warning at {result.path}:{warning.line} {warning.message}")

        if result.content != result.output:
            click.echo(f"Would sort {result.path}")
            return_code |= 2

        print_benchmark(result.timings)

    return return_code


@main.command()
@click.pass_context
@click.argument("filenames", nargs=-1)
@usort_command
def diff(ctx: click.Context, filenames: List[str]) -> int:
    """
    Output diff of changes for one or more path
    """
    if not filenames:
        raise click.ClickException("Provide some filenames")

    return_code = 0
    paths = [Path(f) for f in filenames]
    for result in usort_path(paths, write=False):
        if result.error:
            click.echo(f"Error sorting {result.path}: {result.error}")
            if ctx.obj.debug:
                click.echo(result.trace)
            return_code |= 1
            continue

        for warning in result.warnings:
            click.echo(f"Warning at {result.path}:{warning.line} {warning.message}")

        if result.content != result.output:
            assert result.encoding is not None
            echo_color_unified_diff(
                result.content.decode(result.encoding),
                result.output.decode(result.encoding),
                result.path.as_posix(),
            )

        print_benchmark(result.timings)

    return return_code


@main.command()
@click.argument("filenames", nargs=-1)
@usort_command
def format(filenames: List[str]) -> int:
    """
    Format one or more paths

    This is intended to sort nodes separated by barriers, and that's it.
    We don't format them (aside from moving comments).  Black does the rest.
    When in doubt leave lines alone.
    """
    if not filenames:
        raise click.ClickException("Provide some filenames")

    if filenames[0].strip() == "-":
        success = usort_stdin()
        return 0 if success else 1

    return_code = 0
    paths = [Path(f) for f in filenames]
    for result in usort_path(paths, write=True):
        if result.error:
            click.echo(f"Error sorting {result.path}: {result.error}")
            return_code |= 1
            continue

        for warning in result.warnings:
            click.echo(f"Warning at {result.path}:{warning.line} {warning.message}")

        if result.content != result.output:
            click.echo(f"Sorted {result.path}")

        print_benchmark(result.timings)

    return return_code


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

from usort.tests import *  # noqa: F401,F403

unittest.main()
